"""
This package collects together tools for radio astronomical interferometry.
In addition to pure-python phasing, calibration, imaging, and
deconvolution code, this package includes interfaces to MIRIAD (a Fortran
interferometry package) and HEALPix (a package for representing spherical
data sets), and some math/fitting routines from SciPy.
All code provided is released under the GNU General Public License
(see LICENSE.txt).

Author: Aaron Parsons
"""

from __future__ import print_function, division, absolute_import

from . import phs, const, coord, deconv
from . import fit, healpix, img
from . import interp, cal, map, miriad
import scipy.optimize as optimize
from . import rfi, amp, scripting, src, _src, utils
from . import dsp
from . import pol, twodgauss #added by dfm
import ephem

try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version

try:
    from ._version import version as __version__
except ModuleNotFoundError:  # pragma: no cover
    try:
        __version__ = version("aipy")
    except PackageNotFoundError:
        # package is not installed
        __version__ = "unknown"


"""
The 4C (Fourth Cambridge) Catalog.

Data files are in tab-separated format from Vizier.
To download in the correct format, open a catalog online in Vizier,
select'Tab-Separated-Values' as the Output layout in the drop-down box, set
the maximum entries to 'unlimited', and click 'Sexagesimal' under the box
for 'Target Name or Position'.  Submit the query, and copy the output to a
txt file.  Copy this file to "4c.txt" in the _src directory of your AIPY
installation.
"""

from __future__ import print_function, division, absolute_import

try:
    import aipy as a
except ImportError:
    import aipy as a
import numpy as np, os

class FourCCatalog(a.fit.SrcCatalog):
    def fromfile(self,filename):
        f = open(filename)
        addsrcs = []
        for L in [L for L in f.readlines() if not L.startswith('#')]:
            text = L.split('\t')
            if len(text) <= 4: continue
            try: int(text[0][0])
            except(ValueError): continue
            ra = text[0].replace(' ',':')
            dec = text[1].replace(' ',':')
            name = text[2].strip()
            jys = float(text[5])
            addsrcs.append(a.fit.RadioFixedBody(ra, dec, name=name,
                jys=jys, index=0, mfreq=0.178))
        self.add_srcs(addsrcs)

FOURCFILE = os.path.join(os.path.dirname(__file__), '4c.txt')
_fourccat = None

def get_srcs(srcs=None, cutoff=None):
    global _fourccat
    if _fourccat is None:
        _fourccat = FourCCatalog()
        _fourccat.fromfile(FOURCFILE)
    if srcs is None:
        if cutoff is None: srcs = _fourccat.keys()
        else:
            cut, fq = cutoff
            fq = np.array([fq])
            for s in _fourccat.keys(): _fourccat[s].update_jys(fq)
            srcs = [s for s in _fourccat.keys() if _fourccat[s].jys[0] > cut]
    srclist = []
    for s in srcs:
        try: srclist.append(_fourccat[s])
        except(KeyError): pass
    return srclist



# -*- coding: utf-8 -*-
# Copyright (c) 2011 Aaron Parsons
# Licensed under the GPLv3

import pytest
import aipy.amp as amp
import numpy as np


@pytest.fixture(scope="function")
def generate_AntennaArray():
    freqs = np.arange(0.1, 0.2, 0.01)
    bm = amp.Beam(freqs)
    ant0 = amp.Antenna(0, 0, 0, bm)
    aa = amp.AntennaArray(("0", "0"), [ant0])

    yield freqs, aa

    # clean up when done
    del freqs, aa

    return


@pytest.fixture(scope="function")
def generate_Beam():
    freqs = np.arange(0.1, 0.2, 0.01)
    bm = amp.Beam(freqs)

    yield freqs, bm

    # clean up when done
    del freqs, bm

    return


@pytest.fixture(scope="function")
def generate_Beam2DGaussian():
    freqs = np.arange(0.1, 0.2, 0.01)
    bm = amp.Beam2DGaussian(freqs, 0.05, 0.025)

    yield freqs, bm

    # clean up when done
    del freqs, bm

    return


@pytest.fixture(scope="function")
def generate_BeamAlm():
    freqs = np.arange(0.1, 0.2, 0.01)
    coeffs = {0: np.array([1.0, 0, 0], dtype=np.complex128)}
    bm = amp.BeamAlm(freqs, lmax=1, mmax=1, deg=0, coeffs=coeffs)

    yield freqs, coeffs, bm

    # clean up when done
    del freqs, coeffs, bm

    return


@pytest.fixture(scope="function")
def generate_Antenna():
    freqs = np.arange(0.1, 0.2, 0.01)
    bm = amp.Beam2DGaussian(freqs, 0.05, 0.025)
    ant = amp.Antenna(0, 0, 0, beam=bm)

    yield freqs, ant

    # clean up when done
    del freqs, ant

    return


def test_radiobody_attributes(generate_AntennaArray):
    """Test aipy.amp.RadioFixedBody attributes"""
    freqs, aa = generate_AntennaArray

    source = amp.RadioFixedBody(
        "0:00", "0:00", jys=100, index=-2, mfreq=0.1, name="src1"
    )
    assert source._jys == 100
    assert source.index == -2

    source.compute(aa)
    assert np.allclose(source.get_jys(), 100 * (freqs / 0.1) ** (-2))

    return


def test_response_TestBeam(generate_Beam):
    """Test retrieving aipy.amp.Beam response"""
    freqs, bm = generate_Beam
    xyz = (0, 0, 1)
    assert np.allclose(bm.response(xyz), np.ones_like(freqs))

    x = np.array([0, 0.1, 0.2])
    y = np.array([0.1, 0.2, 0])
    z = np.array([0.2, 0, 0.1])
    xyz = (x, y, z)
    assert np.allclose(bm.response(xyz), np.ones((freqs.size, 3)))

    bm.select_chans([0, 1, 2])
    assert np.allclose(bm.response(xyz), np.ones((3, 3)))

    return


def test_response_TestBeam2DGaussian(generate_Beam2DGaussian):
    """Test retrieving a 2D Gaussian beam response"""
    freqs, bm = generate_Beam2DGaussian
    xyz = (0, 0, 1)
    assert np.allclose(bm.response(xyz), np.ones_like(freqs))

    x = np.array([0, 0.05, 0])
    y = np.array([0, 0, 0.05])
    z = np.array([1, 1, 1])
    xyz = (x, y, z)
    resp = bm.response(xyz)
    assert resp.shape == (freqs.size, 3)

    ans = np.sqrt(np.array([1.0, np.exp(-0.5), np.exp(-2)]))
    ans.shape = (1, 3)
    bm.select_chans([0])
    resp = bm.response(xyz)
    assert np.allclose(np.round(resp - ans, 3), 0.0)

    return


def test_init_update_BeamAlm(generate_BeamAlm):
    freqs, coeffs, bm = generate_BeamAlm
    assert np.allclose(bm.hmap[0].map, 1.0 / np.sqrt(4 * np.pi))

    return


def test_response_BeamAlm(generate_BeamAlm):
    freqs, coeffs, bm = generate_BeamAlm
    x = np.array([0, 0.05, 0])
    y = np.array([0, 0, 0.05])
    z = np.array([1, 1, 1])
    xyz = (x, y, z)
    resp = bm.response(xyz)
    assert resp.shape == (freqs.size, 3)

    ans = np.ones(3, dtype=np.complex128) / np.sqrt(4 * np.pi)
    ans.shape = (1, 3)
    bm.select_chans([0])
    resp = bm.response(xyz)
    assert np.allclose(np.round(resp - ans, 3), 0.0)

    return


def test_passband_Antenna(generate_Antenna):
    """Test the Antenna passband"""
    freqs, ant = generate_Antenna
    pb = ant.passband()
    assert np.allclose(pb, np.ones_like(freqs))

    ant.select_chans([0, 1, 2])
    pb = ant.passband()
    assert pb.shape == (3,)

    return



# -*- coding: utf-8 -*-
# Copyright (c) 2018 Aaron Parsons
# Licensed under the GPLv3

import pytest
import aipy.healpix as ah
import numpy as np


@pytest.fixture(scope="function")
def healpix_base():
    hpb = ah.HealpixBase(nside=32, scheme="RING")
    yield hpb

    # cleanup
    del hpb
    return


@pytest.fixture(scope="function")
def healpix_map():
    hpm = ah.HealpixMap(32, "RING")
    yield hpm

    # cleanup
    del hpm
    return


def test_order(healpix_base):
    """Test HEALpix order functional attribute"""
    hpb = healpix_base
    assert hpb.order() == 5
    return


def test_nside(healpix_base):
    """Test HEALpix nside functional attribute"""
    hpb = healpix_base
    assert hpb.nside() == 32
    return


def test_npix(healpix_base):
    """Test HEALpix npix functional attribute"""
    hpb = healpix_base
    assert hpb.npix() == 12 * hpb.nside() ** 2
    return


def test_scheme(healpix_base):
    """Test HEALpix scheme functional attribute"""
    hpb = healpix_base
    assert hpb.scheme() == "RING"
    return


def test_npix2nside(healpix_base):
    """Test HEALpix npix2nside functional attribute"""
    hpb = healpix_base
    assert hpb.npix2nside(12 * 2 ** 12) == 2 ** 6
    return


def test_set_nside_scheme(healpix_base):
    hpb = healpix_base
    hpb.set_nside_scheme(64, "NEST")
    assert hpb.nside() == 64
    assert hpb.scheme() == "NEST"

    hpb.set_nside_scheme(32, "RING")
    assert hpb.nside() == 32
    assert hpb.scheme() == "RING"
    return


def test_nest_ring_conv(healpix_base):
    hpb = healpix_base
    ipx = np.array([0])
    px = hpb.nest_ring_conv(ipx, "NEST")
    assert px == 1023

    hpb.set_nside_scheme(32, "NEST")
    ipx = np.array([0])
    px = hpb.nest_ring_conv(ipx, "RING")
    assert px == 5968
    return


def test_ang2px(healpix_base):
    hpb = healpix_base
    th = np.linspace(0, np.pi, 10)
    ph = np.linspace(-np.pi, np.pi, 10)
    px = hpb.crd2px(th, ph)
    assert len(px) == len(th)
    np.testing.assert_allclose(
        px, np.array([2, 398, 1375, 3114, 5049, 7239, 9173, 10912, 11889, 12286])
    )
    return


def test_vec2px(healpix_base):
    hpb = healpix_base
    x = np.linspace(-0.5, 0.5, 10)
    y = np.linspace(-0.5, 0.5, 10)
    z = 1 - np.sqrt(x ** 2 + y ** 2)
    px = hpb.crd2px(x, y, z)
    assert len(px) == len(x)
    np.testing.assert_allclose(
        px, np.array([3728, 2192, 1069, 247, 19, 13, 225, 1023, 2128, 3664])
    )
    return


def test_px2vec(healpix_base):
    hpb = healpix_base
    px = np.arange(hpb.npix())
    x, y, z = hpb.px2crd(px)
    assert len(px) == len(x)

    px_recovered = hpb.crd2px(x, y, z)
    np.testing.assert_equal(px, px_recovered)
    return


def test_px2ang(healpix_base):
    hpb = healpix_base
    px = np.arange(hpb.npix())
    th, phi = hpb.px2crd(px, ncrd=2)
    assert len(px) == len(th)

    px_recovered = hpb.crd2px(th, phi)
    np.testing.assert_equal(px, px_recovered)
    return


def test_ang2px_interp(healpix_base):
    hpb = healpix_base
    th = np.linspace(0, np.pi, 3)
    ph = np.linspace(-np.pi, np.pi, 3)
    px, wgt = hpb.crd2px(th, ph, interpolate=True)
    assert len(px) == len(th)
    assert wgt.shape == (len(th), 4)
    np.testing.assert_allclose(
        px,
        np.array(
            [
                [3, 0, 1, 2],
                [6207, 6080, 6208, 6209],
                [12285, 12286, 12287, 12284],
            ]
        ),
    )
    np.testing.assert_allclose(
        wgt,
        np.array(
            [
                [0.25, 0.25, 0.25, 0.25],
                [0.5, 0.5, 0.0, 0.0],
                [0.25, 0.25, 0.25, 0.25],
            ]
        ),
    )
    return


def test_vec2px_interp(healpix_base):
    hpb = healpix_base
    x = np.linspace(-0.5, 0.5, 3)
    y = np.linspace(-0.5, 0.5, 3)
    z = 1 - np.sqrt(x ** 2 + y ** 2)
    px, wgt = hpb.crd2px(x, y, z, interpolate=True)
    assert len(px) == len(x)
    assert wgt.shape == (len(x), 4)
    np.testing.assert_allclose(
        px,
        np.array(
            [
                [3728, 3729, 3855, 3856],
                [1, 2, 3, 0],
                [3664, 3665, 3791, 3792],
            ]
        ),
    )
    np.testing.assert_allclose(
        wgt,
        np.array(
            [
                [0.367711, 0.0, 0.316145, 0.316145],
                [0.25, 0.25, 0.25, 0.25],
                [0.367711, 0.0, 0.316145, 0.316145],
            ]
        ),
        rtol=1e-5,
    )
    return


def test_set(healpix_map):
    hpm = healpix_map
    hpm[0] = 1
    assert hpm[0] == 1

    hpm[0, 0] = 1
    assert hpm[0, 0] == 1

    hpm[0, 1, 0] = 1
    assert hpm[0, 1, 0] == 1
    return


def test_set_px(healpix_map):
    hpm = healpix_map
    cval = np.arange(15)
    assert len(hpm[cval]) == len(cval)

    hpm[cval] += cval
    np.testing.assert_allclose(hpm[cval], cval)

    hpm[cval] += cval
    np.testing.assert_allclose(hpm[cval], 2 * cval)

    hpm[cval] = cval
    np.testing.assert_allclose(hpm[cval], cval)
    return


def test_set_ang(healpix_map):
    hpm = healpix_map
    th = np.linspace(0, np.pi, 10)
    ph = np.linspace(-np.pi, np.pi, 10)
    assert len(hpm[th, ph]) == len(th)

    hpm[th, ph] += th
    np.testing.assert_allclose(hpm[th, ph], th)

    hpm[th, ph] += th
    np.testing.assert_allclose(hpm[th, ph], 2 * th)

    hpm[th, ph] = th
    np.testing.assert_allclose(hpm[th, ph], th)
    return


def test_set_vec(healpix_map):
    hpm = healpix_map
    x = np.linspace(-0.5, 0.5, 10)
    y = np.linspace(-0.5, 0.5, 10)
    z = 1 - np.sqrt(x ** 2 + y ** 2)
    assert len(hpm[x, y, z]) == len(x)

    hpm[x, y, z] += x
    np.testing.assert_allclose(hpm[x, y, z], x)

    hpm[x, y, z] += x
    np.testing.assert_allclose(hpm[x, y, z], 2 * x)

    hpm[x, y, z] = x
    np.testing.assert_allclose(hpm[x, y, z], x)
    return



# -*- coding: utf-8 -*-
# Copyright (c) 2018 Aaron Parsons
# Licensed under the GPLv3

import numpy as np
from aipy import twodgauss


def test_moments():
    """Test twodgauss.moments"""
    a0 = 2.0
    x0 = 5.0
    y0 = 7.0
    wx = 1.0
    wy = 1.0
    x = np.arange(11, dtype=np.float64)
    y = np.arange(11, dtype=np.float64)
    xx, yy = np.meshgrid(x, y)
    data = a0 * np.exp(-(((xx - x0) / wx) ** 2) / 2.0 - ((yy - y0) / wy) ** 2 / 2.0)
    params = twodgauss.moments(data)
    assert np.isclose(a0, params[1], atol=1e-3)
    assert np.isclose(x0, params[2], atol=1e-3)
    assert np.isclose(y0, params[3], atol=1e-3)
    return


def test_twodgaussian():
    """Test twodgauss.twodgaussian"""
    a0 = 2.0
    x0 = 5.0
    y0 = 7.0
    wx = 1.0
    wy = 1.0
    x = np.arange(11, dtype=np.float64)
    y = np.arange(11, dtype=np.float64)
    xx, yy = np.meshgrid(x, y)
    data0 = a0 * np.exp(-(((xx - x0) / wx) ** 2) / 2.0 - ((yy - y0) / wy) ** 2 / 2.0)
    data1 = twodgauss.twodgaussian([0.0, a0, x0, y0, wx, wy], shape=yy.shape)
    assert np.allclose(data0, data1, atol=1e-6)
    return



# -*- coding: utf-8 -*-
# Copyright (c) 2010 Aaron Parsons
# Licensed under the GPLv3

import pytest
import numpy as np

from aipy import miriad
from aipy import _miriad


@pytest.fixture(scope="function")
def test_file_r(tmp_path):
    filename1 = str(tmp_path / "test1.uv")
    filename2 = str(tmp_path / "test2.uv")
    uv = miriad.UV(filename1, status="new", corrmode="r")
    uv["history"] = "Made this file from scratch.\n"
    uv.add_var("nchan", "i")
    uv.add_var("pol", "i")
    uv["nchan"] = 4
    uv["pol"] = -5
    uvw = np.array([1, 2, 3], dtype=np.float64)
    preamble = (uvw, 12345.6789, (0, 1))
    data = np.ma.array([1j, 2, 3j, 4], mask=[0, 0, 1, 0], dtype=np.complex64)
    uv.write(preamble, data)
    uv["pol"] = -6
    uv.write(preamble, data)
    del uv

    yield filename1, filename2, data

    return


@pytest.fixture(scope="function")
def test_file_j(tmp_path):
    filename = str(tmp_path / "test1.uv")
    uv = miriad.UV(filename, status="new", corrmode="j")
    uv["history"] = "Made this file from scratch.\n"
    uv.add_var("nchan", "i")
    uv.add_var("pol", "i")
    uv["nchan"] = 4
    uv["pol"] = -5
    uvw = np.array([1, 2, 3], dtype=np.float64)
    preamble = (uvw, 12345.6789, (0, 1))
    data = np.ma.array([1j, 2, 3j, 4], mask=[0, 0, 1, 0], dtype=np.complex64)
    uv.write(preamble, data)
    uv["pol"] = -6
    uv.write(preamble, data)
    del uv

    yield filename, data

    return


def test_maxchan():
    assert type(_miriad.MAXCHAN) == int
    return


def test_immediate_corr(test_file_r):
    """Test immediate corr of a Miriad UV file"""
    filename1, filename2, data = test_file_r
    uv = miriad.UV(filename2, status="new")
    assert uv.vartable["corr"] == "r"
    return


def test_vartable_r(test_file_r):
    """Test accesing vartable data in a Miriad UV file"""
    filename1, filename2, data = test_file_r
    uv1 = miriad.UV(filename1)
    assert uv1.vartable["corr"] == "r"
    assert uv1.vartable["nchan"] == "i"
    assert uv1.vartable["pol"] == "i"
    return


def test_data_r(test_file_r):
    """Test writing data from a Miriad UV file"""
    filename1, filename2, data = test_file_r
    uv = miriad.UV(filename1)
    assert uv["history"] == "Made this file from scratch.\n"
    (uvw, t, bl), d = uv.read()
    assert uv["nchan"] == 4
    assert uv["pol"] == -5
    assert bl == (0, 1)
    assert np.isclose(t, 12345.6789)
    assert np.allclose(uvw, np.array([1, 2, 3], dtype=np.float64))
    assert np.allclose(d, data)

    (uvw, t, bl), d = uv.read()
    assert uv["nchan"] == 4
    assert uv["pol"] == -6
    assert bl == (0, 1)
    assert np.isclose(t, 12345.6789)
    assert np.allclose(uvw, np.array([1, 2, 3], dtype=np.float64))
    assert np.allclose(d, data)
    return


def test_vartable_j(test_file_j):
    """Test accesing vartable data in a Miriad UV file"""
    filename, data = test_file_j
    uv = miriad.UV(filename, corrmode="j")
    assert uv.vartable["corr"] == "j"
    assert uv.vartable["nchan"] == "i"
    assert uv.vartable["pol"] == "i"
    return


def test_data_j(test_file_j):
    """Test writing data from a Miriad UV file"""
    filename, data = test_file_j
    uv = miriad.UV(filename)
    assert uv["history"] == "Made this file from scratch.\n"

    (uvw, t, bl), d = uv.read()
    assert uv["nchan"] == 4
    assert uv["pol"] == -5
    assert bl == (0, 1)
    assert np.isclose(t, 12345.6789)
    assert np.allclose(uvw, np.array([1, 2, 3], dtype=np.float64))
    assert np.allclose(np.abs(d - data), 0.0, atol=1e-4)

    (uvw, t, bl), d = uv.read()
    assert uv["nchan"] == 4
    assert uv["pol"] == -6
    assert bl == (0, 1)
    assert np.isclose(t, 12345.6789)
    assert np.allclose(uvw, np.array([1, 2, 3], dtype=np.float64))
    assert np.allclose(np.abs(d - data), 0.0, atol=1e-4)
    return



# -*- coding: utf-8 -*-
# Copyright (c) 2008 Aaron Parsons
# Licensed under the GPLv3

import pytest
import numpy as np
import aipy


@pytest.fixture(scope="function")
def init_deconv():
    SIZE = 100
    NOISE = 0.001
    img = np.zeros((SIZE, SIZE), np.float64)
    img[10, 10] = 10
    img[20:25, 20:25] = 1
    img[30:40, 30:40] = 0.1

    bm = aipy.img.gaussian_beam(2, shape=img.shape)
    bm[0, 0] = 0.05

    data = np.abs(np.fft.ifft2(np.fft.fft2(img) * np.fft.fft2(bm)))
    ns = np.random.normal(scale=NOISE, size=img.shape)
    data = np.abs(data + ns)

    yield data, bm

    # clean up when done
    del data, bm

    return


def test_clean(init_deconv):
    """Test that the standard clean deconvolution runs"""
    data, bm = init_deconv
    cln, info = aipy.deconv.clean(data, bm, verbose=False)
    assert True

    return


def test_lsq(init_deconv):
    """Test that least squared deconvolution runs"""
    data, bm = init_deconv
    cln, info = aipy.deconv.lsq(data, bm, verbose=False)
    assert True

    return


def test_mem(init_deconv):
    """Test the maximum entropy deconvolution runs"""
    data, bm = init_deconv
    cln, info = aipy.deconv.maxent(data, bm, np.var(data ** 2) * 0.5, verbose=False)
    assert True

    return


def test_anneal(init_deconv):
    """Test that simulated annealing deconvolution runs"""
    data, bm = init_deconv
    cln, info = aipy.deconv.anneal(data, bm, verbose=False)
    assert True

    return



#!/usr/bin/env python

"""
Helps you figure out which version of AIPY you have installed.
Lists:
    version
    install location
    last commit
    branch
"""

from __future__ import print_function, division, absolute_import

import aipy
print("AIPY base version in use: ", end='')
try:
    print(aipy.__version__)
except(AttributeError):
    print("Version number not Found")
print("Install location:", aipy.__file__)
print("Last Git commit log:")
try:
    print(aipy.__gitlog__)
except(AttributeError):
    print("not found")
print("Branch: ", end='')
try:
    print(aipy.__branch__)
except(AttributeError):
    print("unknown")



# -*- coding: utf-8 -*-
# Copyright (c) 2008 Aaron Parsons
# Licensed under the GPLv3

import random

import aipy
import pytest
import ephem
import numpy as np


@pytest.fixture(scope="function")
def source_catalog():
    src1 = aipy.phs.RadioFixedBody("1:00", "1:00", name="src1")
    src2 = aipy.phs.RadioFixedBody("2:00", "2:00", name="src2")
    src3 = aipy.phs.RadioFixedBody("3:00", "3:00", name="src3")
    srcs = [src1, src2, src3]
    cat = aipy.phs.SrcCatalog(srcs)
    yield srcs, cat
    return


@pytest.fixture(scope="function")
def test_beam():
    freqs = np.arange(0, 1, 0.1)
    bm = aipy.phs.Beam(freqs)
    yield freqs, bm
    return


@pytest.fixture(scope="function")
def test_antenna():
    freqs = np.arange(0, 1, 0.1)
    bm = aipy.phs.Beam(freqs)
    ant = aipy.phs.Antenna(1, 2, 3, bm, phsoff=[0, 1])
    yield freqs, bm, ant
    return


@pytest.fixture(scope="function")
def test_array_location():
    yield aipy.phs.ArrayLocation(("0", "0"))
    return


@pytest.fixture(scope="function")
def test_antenna_array():
    bm = aipy.phs.Beam(np.arange(0, 1, 0.1))
    a1 = aipy.phs.Antenna(0, 0, 0, bm, [0, 0])
    a2 = aipy.phs.Antenna(1, 0, 0, bm, [0, 1])
    a3 = aipy.phs.Antenna(0, 1, 0, bm, [0, 2])
    a4 = aipy.phs.Antenna(0, 0, 1, bm, [0, 3])
    ants = [a1, a2, a3, a4]
    aa = aipy.phs.AntennaArray(("0", "0"), ants)
    yield ants, aa
    return


def test_pointing_error():
    pnterr = aipy.phs.PointingError("Error String")
    assert str(pnterr) == "Error String"
    with pytest.raises(aipy.phs.PointingError):
        raise pnterr
    return


def test_ephem_zero():
    """Test converting between ephem dates and JD - zero point"""
    ephemzero = ephem.date("1899/12/31 12:00")
    jdzero = 2415020.0
    assert np.isclose(aipy.phs.juldate2ephem(jdzero), ephemzero)
    assert np.isclose(aipy.phs.ephem2juldate(ephemzero), jdzero)
    return


def test_ephem_random():
    """Test converting between ephem dates and JD - various"""
    for i in range(10):
        d1 = random.random() * ephem.now()
        d2 = aipy.phs.juldate2ephem(aipy.phs.ephem2juldate(d1))
        assert np.isclose(d1, d2)

    return


def test_radiobody_attributes():
    """Test aipy.phs.RadioFixedBody attributes"""
    epoch = ephem.B1950
    source = aipy.phs.RadioFixedBody(
        "0:00",
        "0:00",
        mfreq=0.200,
        name="src1",
        epoch=epoch,
        ionref=(0.0, 0.0),
        srcshape=(0.003, 0.005, 0.6),
    )
    assert source._ra == ephem.hours("0:00")
    assert source._dec == ephem.degrees("0:00")
    assert np.isclose(source.mfreq, 0.200)
    assert source.src_name == "src1"
    assert np.isclose(source._epoch, epoch, atol=1e-3)
    assert np.allclose(source.ionref, np.array([0.0, 0.0]))
    assert source.srcshape == [0.003, 0.005, 0.6]

    with pytest.raises(RuntimeError):
        source.ra
    with pytest.raises(RuntimeError):
        source.dec
    with pytest.raises(AttributeError):
        source.map

    obs = aipy.phs.ArrayLocation(("0:00", "0:00"))
    obs.set_ephemtime(epoch)
    source.compute(obs)
    assert len(source.get_crds("eq", ncrd=2)) == 2
    assert len(source.get_crds("top", ncrd=2)) == 2
    assert len(source.get_crds("eq", ncrd=3)) == 3
    assert len(source.get_crds("top", ncrd=3)) == 3
    assert source.map.shape == (3, 3)
    return


def test_ephem_interface():
    """Test the aipy.phs.RadioFixedBody ephem interface"""
    obs = aipy.phs.ArrayLocation(("0:00", "0:00"))
    for epoch in [ephem.B1900, ephem.B1950, ephem.J2000]:
        for ra in np.arange(np.pi / 8, 2 * np.pi, np.pi / 8):
            for dec in np.arange(-3 * np.pi / 8, np.pi / 2, np.pi / 8):
                source = aipy.phs.RadioFixedBody(ra, dec, epoch=epoch)
                obs.set_ephemtime(epoch)
                source.compute(obs)
                assert np.isclose(source.a_ra, source._ra, atol=1e-9)
                assert np.isclose(source.a_dec, source._dec, atol=1e-9)
    return


def test_ephem_compute():
    """Test the aipy.phs.RadioFixedBody ephem calculations"""
    epoch = ephem.J2000
    obs = aipy.phs.ArrayLocation(("0:00", "0:00"))
    obs.set_ephemtime(epoch)
    diagonal = np.array([[1.0, 0, 0], [0, 1, 0], [0, 0, 1]])
    for ra in np.arange(np.pi / 8, 2 * np.pi, np.pi / 8):
        for dec in np.arange(-3 * np.pi / 8, np.pi / 2, np.pi / 8):
            source = aipy.phs.RadioFixedBody(ra, dec, epoch=epoch)
            source.compute(obs)
            mat = aipy.coord.top2eq_m(obs.sidereal_time() - source.ra, source.dec)
            err = np.abs(diagonal - np.dot(mat, source.map)).sum()
            assert np.isclose(err, 0.0, atol=1e-10)
    return


def test_get_crds(self):
    """Test the aipy.phs.RadioFixedBody calculated coordinates"""
    epoch = ephem.J2000
    obs = aipy.phs.ArrayLocation(("0:00", "0:00"))
    obs.set_ephemtime(epoch)
    for ra in np.arange(np.pi / 8, 2 * np.pi, np.pi / 8):
        for dec in np.arange(-3 * np.pi / 8, np.pi / 2, np.pi / 8):
            source = aipy.phs.RadioFixedBody(ra, dec, epoch=epoch)
            source.compute(obs)
            ra, dec = source.get_crds("eq", ncrd=2)
            assert np.isclose(source.ra, ra)
            assert np.isclose(source.dec, dec)

            az, alt = source.get_crds("top", ncrd=2)
            assert np.isclose(source.az, az)
            assert np.isclose(source.alt, alt)

            eq = source.get_crds("eq", ncrd=3)
            ra, dec = aipy.coord.eq2radec(eq)
            assert np.isclose(source.ra, ra, atol=1e-10)
            assert np.isclose(source.dec, dec, atol=1e-10)

            top = source.get_crds("top", ncrd=3)
            az, alt = aipy.coord.top2azalt(top)
            assert np.isclose(source.az, az, atol=1e-10)
            assert np.isclose(source.alt, alt, atol=1e-10)
    return


def test_add_srcs(source_catalog):
    """Test adding sources to a aipy.phs.SrcCatalog() catalog"""
    srcs, cat = source_catalog
    cat2 = aipy.phs.SrcCatalog()
    src1b = aipy.phs.RadioFixedBody("0:00", "0:00", name="src1")
    src4 = aipy.phs.RadioFixedBody("4:00", "4:00", name="src4")
    cat2.add_srcs(srcs)
    assert len(cat2) == 3
    cat2.add_srcs(src1b)
    assert len(cat2) == 3
    cat2.add_srcs(src1b, src4)
    assert len(cat2) == 4

    srclist = [src1b] + srcs[1:] + [src4]
    for name, src in zip([s.src_name for s in srclist], srclist):
        assert src == cat2[name]
    return


def test_get_srcs(source_catalog):
    """Test retrieving sources from a aipy.phs.SrcCatalog() catalog"""
    srcs, cat = source_catalog
    assert cat.get_srcs("src1", "src2", "src3") == srcs
    assert cat.get_srcs(["src1", "src2"]) == srcs[:2]
    with pytest.raises(KeyError):
        cat.get_srcs("bad")
    return


def test_compute(source_catalog):
    """Test the ephem interfaces for a aipy.phs.SrcCatalog() catalog"""
    srcs, cat = source_catalog
    obs = ephem.Observer()
    cat.compute(obs)
    for src in cat.values():
        assert src.ra is not None
        assert src.dec is not None
    return


def test_get_crds(source_catalog):
    """Test coordinates calculated from a aipy.phs.SrcCatalog() catalog"""
    srcs, cat = source_catalog
    obs = ephem.Observer()
    cat.compute(obs)
    crd1 = cat.get_crds("eq", srcs=["src1"])
    assert crd1.shape == (3, 1)
    assert np.allclose(crd1[:, 0], srcs[0].get_crds("eq"))

    crd2 = cat.get_crds("top", srcs=["src1", "src2"])
    assert crd2.shape == (3, 2)
    return


def test_get(source_catalog):
    """Test retrieving source attributes from a aipy.phs.SrcCatalog() catalog"""
    srcs, cat = source_catalog
    mfreq = cat.get("mfreq", srcs=["src1"])
    assert mfreq.shape == (1,)

    mfreq = cat.get("mfreq", srcs=["src1", "src2"])
    assert mfreq.shape == (2,)

    ionrefs = cat.get("ionref", srcs=["src1", "src2"])
    assert ionrefs.shape == (2, 2)

    srcshapes = cat.get("srcshape", srcs=["src1", "src2"])
    assert srcshapes.shape == (3, 2)
    return


def test_beam_attributes(test_beam):
    """Test accessing aipy.phs.Beam attributes"""
    freqs, bm = test_beam
    assert np.allclose(bm.freqs, freqs)
    assert np.allclose(bm.chans, np.arange(freqs.size))
    assert np.allclose(bm.afreqs, freqs)
    return


def test_beam_select_chans(test_beam):
    """Test selecting various aipy.phs.Beam channels"""
    freqs, bm = test_beam
    chans = np.array([1, 2, 3])
    bm.select_chans(chans)
    assert np.allclose(bm.chans, chans)
    assert np.allclose(bm.afreqs, freqs.take(chans))
    return


def test_antenna_attributes(test_antenna):
    """Test accessing aipy.phs.Antenna attributes"""
    freqs, bm, ant = test_antenna
    assert ant.beam == bm
    pos = np.array([1, 2, 3], dtype=np.float64)
    assert np.allclose(pos, ant.pos)
    x, y, z = ant
    assert np.isclose(x, pos[0])
    assert np.isclose(y, pos[1])
    assert np.isclose(z, pos[2])
    assert np.allclose(ant + ant, pos * 2)
    assert np.allclose(ant - ant, 0.0)
    assert np.allclose(-ant, -pos)
    return


def test_antenna_select_chans(test_antenna):
    """Test selecting various aipy.phs.Antenna channels"""
    freqs, bm, ant = test_antenna
    chans = np.array([1, 2, 3])
    ant.select_chans(chans)
    assert np.allclose(bm.chans, chans)
    assert np.allclose(ant.phsoff, 1)
    return


def test_array_location_attributes(test_array_location):
    al = test_array_location
    assert al.pressure == 0.0
    assert al.lat == 0.0
    assert al.long == 0.0
    assert al.elev == 0.0
    assert np.allclose(al._eq2zen, aipy.coord.eq2top_m(0.0, 0.0))
    return


def test_array_location_set_jultime(test_array_location):
    al = test_array_location
    jd = 2454554.9798841
    al.set_jultime(jd)
    assert al.date == al.epoch
    assert al.date == aipy.phs.juldate2ephem(jd)
    assert np.isclose(al.sidereal_time(), 0, atol=1e-7)

    eq2now_rnd = np.round(al._eq2now, 7)
    assert np.allclose(
        eq2now_rnd,
        np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]),
    )
    return


def test_array_location_get_jultime(test_array_location):
    al = test_array_location
    al.set_jultime(2454555)
    assert al.get_jultime() == 2454555
    return


def test_antenna_array_attributes(test_antenna_array):
    ants, aa = test_antenna_array
    assert len(aa) == 4
    for ai, aj in zip(aa, ants):
        assert ai == aj
    for i in range(4):
        assert len(aa[:i]) == i
    return


def test_antenna_array_select_chans(test_antenna_array):
    ants, aa = test_antenna_array
    chans = np.array([1, 2, 3])
    aa.select_chans(chans)
    for ant in aa.ants:
        assert np.allclose(ant.beam.chans, chans)
    return


def test_antenna_array_ij2bl(test_antenna_array):
    ants, aa = test_antenna_array
    assert aa.ij2bl(0, 1) == 258
    assert aa.ij2bl(0, 2) == 259
    assert aa.ij2bl(1, 2) == 515
    return


def test_antenna_array_bl2ij(test_antenna_array):
    ants, aa = test_antenna_array
    assert aa.bl2ij(258) == (0, 1)
    assert aa.bl2ij(259) == (0, 2)
    assert aa.bl2ij(515) == (1, 2)
    return


def test_antenna_array_get_baseline(test_antenna_array):
    ants, aa = test_antenna_array
    for j, ant in enumerate(aa):
        if j in [0, 3]:
            aa.set_jultime(2454554.9)
            assert np.allclose(aa.get_baseline(0, j, "r"), ant.pos)
            assert np.allclose(aa.get_baseline(0, j, "e"), ant.pos)
        else:
            assert np.allclose(aa.get_baseline(0, j, "r"), ant.pos)
            aa.set_jultime(2454554.9798841)
            bl_rnd = np.round(aa.get_baseline(0, j, "e"), 7)
            assert np.allclose(bl_rnd, ant.pos)

            aa.set_jultime(2454554.9)
            bl_rnd = np.round(aa.get_baseline(0, j, "e"), 7)
            assert not np.allclose(bl_rnd, ant.pos)
    src = aipy.phs.RadioFixedBody("12:00", "0:00")
    src.compute(aa)
    with pytest.raises(aipy.phs.PointingError):
        aa.get_baseline(0, 1, src)
    for t in np.random.random((10,)):
        aa.set_jultime(2454555.0 + t)
        src = aipy.phs.RadioFixedBody(aa.sidereal_time(), aa.lat, epoch=aa.epoch)
        src.compute(aa)
        zbl_rnd = np.round(aa.get_baseline(0, 1, "z"), 3)
        sbl_rnd = np.round(aa.get_baseline(0, 1, src), 3)
        assert np.allclose(zbl_rnd, sbl_rnd)
    return


def test_antenna_array_get_phs_offset(test_antenna_array):
    ants, aa = test_antenna_array
    assert np.allclose(aa.get_phs_offset(0, 0), 0)
    assert np.allclose(aa.get_phs_offset(0, 1), 1)
    assert np.allclose(aa.get_phs_offset(0, 2), 2)
    assert np.allclose(aa.get_phs_offset(0, 3), 3)
    return


def test_antenna_array_gen_uvw(test_antenna_array):
    ants, aa = test_antenna_array
    aa.select_chans()
    afreqs = aa[0].beam.afreqs
    u, v, w = aa.gen_uvw(0, 1, "z")
    assert u.shape == (1, afreqs.size)
    assert np.allclose(u, 0 * afreqs)
    assert np.allclose(v, 0 * afreqs)
    assert np.allclose(w, 1 * afreqs)

    u, v, w = aa.gen_uvw(0, 2, "z")
    assert np.allclose(u, 1 * afreqs)
    assert np.allclose(v, 0 * afreqs)
    assert np.allclose(w, 0 * afreqs)

    u, v, w = aa.gen_uvw(0, 3, "z")
    assert np.allclose(u, 0 * afreqs)
    assert np.allclose(v, 1 * afreqs)
    assert np.allclose(w, 0 * afreqs)
    return


def test_antenna_array_gen_phs(test_antenna_array):
    ants, aa = test_antenna_array
    aa.select_chans([1, 2, 3])
    afreqs = aa[0].beam.afreqs
    for t in np.random.random((40,)):
        aa.set_jultime(2454555.0 + t)
        src = aipy.phs.RadioFixedBody(aa.sidereal_time(), aa.lat, epoch=aa.epoch)
        src.compute(aa)
        if t > 0.5:
            seq = src.get_crds("eq", ncrd=3)
            if t > 0.75:
                seq = np.array([seq, seq]).transpose()
        else:
            seq = src
        phs = np.round(aa.gen_phs(seq, 0, 1, mfreq=0.1), 6)
        ans = np.round(np.exp(-1j * 2 * np.pi * afreqs), 6)
        if t > 0.75:
            assert phs.shape == (2, 3)
        else:
            assert phs.shape == (3,)
        assert np.allclose(phs, ans)
        phs = np.round(aa.gen_phs(src, 0, 2, mfreq=0.1), 3)
        assert np.allclose(phs, 1 + 0j)
        phs = np.round(aa.gen_phs(src, 0, 3, mfreq=0.1), 3)
        assert np.allclose(phs, 1 + 0j)

    phs1 = aa.gen_phs(src, 0, 2, mfreq=0.1, ionref=(0.001, 0.001))
    phs2 = aa.gen_phs(src, 0, 2, mfreq=0.1, srcshape=(0.01, 0.01, 0), resolve_src=True)
    assert np.all(phs1 != 1 + 0j)
    assert np.all(phs2 != 1 + 0j)
    return


def test_antenna_array_resolve_src(test_antenna_array):
    ants, aa = test_antenna_array
    # we get a runtime warning when x ~ 0
    with pytest.warns(RuntimeWarning, match="invalid value encountered in") as record:
        amp = aa.resolve_src(100.0, 100.0, srcshape=(0, 0, 0))
    assert len(record) == 1

    # check that value is correct
    assert amp == 1

    amp1 = aa.resolve_src(100.0, 50.0, srcshape=(0.01, 0, np.pi / 2))
    amp2 = aa.resolve_src(100.0, 50.0, srcshape=(0, 0.01, 0))
    assert np.isclose(amp1, amp2, atol=1e-15)

    amp1 = aa.resolve_src(100.0, 100.0, srcshape=(0.02, 0, 0))
    amp2 = aa.resolve_src(100.0, 100.0, srcshape=(0.01414, 0.01414, 0))
    assert np.isclose(amp1, amp2, atol=1e-3)

    amp = aa.resolve_src(100.0, 0.0, srcshape=(0.001, 0, 0))
    x = 2 * np.pi * 0.1
    assert np.isclose(amp, 2 * aipy.phs.j1(x) / x)
    return


def test_refract(test_antenna_array):
    ants, aa = test_antenna_array
    aa.select_chans([1, 2, 3])
    afreqs = aa[0].beam.afreqs
    zeros = np.zeros((1, 3), dtype=np.float64)
    ones = np.ones((1, 3), dtype=np.float64)
    # Test non-vectors, dra->u association
    dw = aa.refract(ones, zeros, mfreq=0.1, ionref=(0.001, 0))
    ans = 0.001 / (afreqs / 0.1) ** 2
    ans.shape = (1, ans.size)
    assert len(dw.shape) == 2
    assert np.allclose(np.round(dw, 10), np.round(ans, 10))

    # Test non-vectors, no dra-> v association
    dw = aa.refract(zeros, ones, mfreq=0.1, ionref=(0.001, 0))
    assert np.allclose(dw, 0)

    # Test non-vectors, no ddec->u association
    dw = aa.refract(ones, zeros, mfreq=0.1, ionref=(0, 0.001))
    assert np.allclose(dw, 0)

    # Test non-vectors, ddec->v association, v scaling
    dw = aa.refract(zeros, 2 * ones, mfreq=0.1, ionref=(0, 0.001))
    assert np.allclose(np.round(dw, 10), np.round(2 * ans, 10))

    # Test vectors, mfreq scaling
    ones = np.ones((2, 3), dtype=np.float64)
    ionref = (np.array([0, 0.001]), np.array([0.001, 0]))
    mfreq = np.array([0.1, 0.2])
    ans = np.array([0.002 / (afreqs / 0.1) ** 2, 0.002 / (afreqs / 0.2) ** 2])
    dw = aa.refract(2 * ones, 2 * ones, mfreq=mfreq, ionref=ionref)
    assert np.allclose(np.round(dw, 10), np.round(ans, 10))
    return


def test_antenna_array_phs2src(test_antenna_array):
    ants, aa = test_antenna_array
    aa.select_chans([1, 2, 3])
    aa.set_jultime(2454555.0)
    src = aipy.phs.RadioFixedBody("0:00", "20:00")
    src.compute(aa)
    assert np.allclose(aa.phs2src(1.0, src, 0, 1), aa.gen_phs(src, 0, 1))
    return


def test_unphs2src(test_antenna_array):
    ants, aa = test_antenna_array
    aa.select_chans([1, 2, 3])
    aa.set_jultime(2454555.0)
    src = aipy.phs.RadioFixedBody("0:00", "20:00")
    src.compute(aa)
    assert np.allclose(aa.unphs2src(aa.gen_phs(src, 0, 1), src, 0, 1), 1.0)
    return



from __future__ import print_function, division, absolute_import

import unittest, aipy._src.four_c as h, aipy as a, numpy as n

class TestFourCCatalog(unittest.TestCase):
    def setUp(self):
        self.cat = h.FourCCatalog()
        self.cat.fromfile(h.FOURCFILE)
    def test_spotcheck(self):
        for srcname in self.cat:
            src = self.cat[srcname]
            self.assertEqual(src.index, 0)
            self.assertEqual(len(src.srcshape), 3)
            self.assertEqual(src.srcshape[0], 0)
            self.assertEqual(src.srcshape[1], 0)
            self.assertEqual(src.srcshape[2], 0)
            self.assertEqual(src.mfreq, .178)

if __name__ == '__main__':
    unittest.main()
